package pgdp.streams;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

public class StreamTemperatures extends Temperatures {
	
	// Temperatures stellt eine stream() Methode bereit!
	
    public StreamTemperatures(String filepath) {
        super(filepath);
    }

    public long size() {
        return stream()
        		//Elemente mit Count Element zählen
        		.count();
    }

    public List<LocalDate> dates() {
        return stream()
        		// Temperaturdaten auf Datum mappen
        		// .map(temp -> temp.getDate) mit Lamda Expression
        		.map(Temperature::getDate)
        		// Duplikate eliminieren
        		.distinct()
        		// sortieren (Comparator ist bereits definiert, da LocalDate Comparable implementiert)
        		.sorted()
        		// mit Collector zur List zusammenfassen
        		.collect(Collectors.toList());
    }

    public Set<String> cities() {
        return stream()
        		// Temperaturdaten auf Städte mappen
        		// .map(temp -> temp.getCity) mit Lamda Expression
        		.map(Temperature::getCity)
        		// mit Collector zu Set (keine Duplikate) zusammenfassen
        		.collect(Collectors.toSet());
    }

    public Set<String> countries() {
    	return stream()
        		// Temperaturdaten auf Länder mappen
        		// .map(temp -> temp.getCountry) mit Lamda Expression
        		.map(Temperature::getCountry)
        		// mit Collector zu Set (keine Duplikate) zusammenfassen
        		.collect(Collectors.toSet());
    }

    public Map<String, List<Double>> temperaturesByCountry() {
    	// Map erstellen, muss final sein, damit Streams damit arbeiten können
    	final Map<String, List<Double>> countryTemperatures = new HashMap<>();
    	
    	// für jedes Land eine Liste von Doubles in der Map anlegen
        countries().forEach(country -> countryTemperatures.put(country,
                new LinkedList<Double>()));
        
        // Temperaturen der jeweiligen Liste hinzufügen
        stream().forEach(
                temperature -> countryTemperatures.get(temperature.getCountry())
                        .add(temperature.getTemperature()));
        return countryTemperatures;
        

    }

    @Override
    public String coldestCountryAbs() {
        return stream()
        		// minimales Element herausfinden
        		.min(Comparator.comparing(Temperature::getTemperature))
        		// auf dem einen Element get() ausführen, um es zu bekommen
                .get()
                // Land zauslesen
                .getCountry();
    }

    @Override
    public String hottestCountryAbs() {
    	 return stream()
         		// maximales Element herausfinden
         		.max(Comparator.comparing(Temperature::getTemperature))
         		// auf dem einen Element get() ausführen, um es zu bekommen
                 .get()
                 // Land zauslesen
                 .getCountry();
    }

    @Override
    public Map<String, Double> countriesAvgTemperature() {
    	// neue Map erstellen, muss final sein, damit Streams damit arbeiten können
    	final Map<String, Double> countryAvgTemperatures = new HashMap<>();
    	
    	// Aus der alten Map mit den Temperaturen pro Land alle Einträge als Set auslesen
        temperaturesByCountry().entrySet().stream()
        
        // jedes Land der neuen Map hinzufügen und durchschnittstemperatur berechnen
                .forEach(entry -> countryAvgTemperatures.put(entry.getKey(),
                // –––––––––––––––––––––––––––––––––––––––––– ^ Ländernamen hinzufügen
                		
                		// Durchschnittstemperatur hinzufügen, indem man Liste zum Stream umwandelt und average Funktion benutzt
                        entry.getValue().stream()
                                .mapToDouble(Double::doubleValue)
                                .average()
                                .getAsDouble()
                                )
                		);
        
        return countryAvgTemperatures;
    }

    
    public Map<String, Double> avgTemperatureDeltaPerYearPerCountry() {
    	// Vorbereitungen: Map erstellen!
    	// Map für String und Liste von Temperaturen
    	 Map<String, List<Temperature>> tempsPerCountry = stream()
                 .collect(Collectors.groupingBy(Temperature::getCountry));
    	 
    	 
         Map<String, Double> avgTempDeltaPerCountry = tempsPerCountry.entrySet()
        		 // für jedes Land Stream von Temperaturen erstellen
        		 .stream().collect(Collectors.toMap(Entry::getKey, entry -> {
        			 
        			 // Map mit Temperaturmessungen pro Jahr (immer nur für aktulles Land 'entry')
                     Map<Integer, List<Temperature>> tempsPerYear = entry
                             .getValue().stream()
                             .collect(Collectors
                                     .groupingBy(temperature -> temperature
                                             .getDate().getYear()));
                     
                     // Durchschnittstemperatur pro Jahr berechnen
                     double[] avgTempsPerYear = tempsPerYear.entrySet().stream()
                    		 // nach Jahr sortieren, damit nachher ein Delta zwischen zwei aufeinanderfolgenden Jahren berechnet werden kann
                             .sorted(Comparator
                                     .comparing(entry2 -> entry2.getKey()))
                             // Durchschnitt berechnen...
                             .mapToDouble(entry2 -> entry2.getValue().stream()
                                     .mapToDouble(temp -> temp.getTemperature())
                                     .average().getAsDouble())
                             // ...und in Array packen
                             .toArray();
                     
                     // Delta für Array index 0 bis länge-2 (range ist exklusive letztes Element)
                     return IntStream.range(0, avgTempsPerYear.length - 1)
                             .mapToDouble(i -> avgTempsPerYear[i + 1]
                                     - avgTempsPerYear[i])
                             // durschnittliches Delta über die Jahre berechnen
                             .average().getAsDouble();
                 }));
         // diese Lamda Expression wird für jedes Land ausgeführt
         
         // globales delta hinzufügen
         avgTempDeltaPerCountry.put("Globally", avgTempDeltaPerCountry.values()
                 .stream().mapToDouble(it -> it).average().getAsDouble());
         return avgTempDeltaPerCountry;
    }

    
    public static void main(final String[] args) {
        String filepath = "temperaturesEurope1963Till2013ByCity.csv";
        StreamTemperatures temperatures = new StreamTemperatures(filepath);

        temperatures.printSummary();
        final Map<String, Double> values = temperatures
                .avgTemperatureDeltaPerYearPerCountry();

        print("Averaged yearly temperature delta per country:",
                Arrays.toString(values.entrySet().toArray()));
    }

}